Class: PgHaMigrations::PartmanConfig
- Inherits:
-
ActiveRecord::Base
- Object
- ActiveRecord::Base
- PgHaMigrations::PartmanConfig
- Defined in:
- lib/pg_ha_migrations/partman_config.rb
Overview
This is an internal class that is not meant to be used directly
Class Method Summary collapse
-
.schema=(schema) ⇒ Object
This method is called by unsafe_partman_update_config to set the fully qualified table name, as partman is often installed in a schema that is not included the application’s search path.
Class Method Details
.schema=(schema) ⇒ Object
This method is called by unsafe_partman_update_config to set the fully qualified table name, as partman is often installed in a schema that is not included the application’s search path
8 9 10 |
# File 'lib/pg_ha_migrations/partman_config.rb', line 8 def self.schema=(schema) self.table_name = "#{schema}.part_config" end |